/*  RAxML-VI-HPC (version 2.2) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright August 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein. 
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  Alexandros.Stamatakis@epfl.ch
 *
 *  When publishing work that is based on the results from RAxML-VI-HPC please cite:
 *
 *  Alexandros Stamatakis:"RAxML-VI-HPC: maximum likelihood-based phylogenetic analyses with thousands of taxa and mixed models". 
 *  Bioinformatics 2006; doi: 10.1093/bioinformatics/btl446
 */

#ifndef WIN32 
#include <unistd.h>
#endif

#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "axml.h"


#ifdef __SIM_SSE3
#include <xmmintrin.h>
#include <pmmintrin.h>
/*#include <tmmintrin.h>*/
#endif

#ifdef _USE_PTHREADS
extern volatile double *reductionBuffer;
extern volatile int NumberOfThreads;
#endif



void calcDiagptable(double z, int data, int numberOfCategories, double *rptr, double *EIGN, double *diagptable)
{
  int i, l;
  double lz;

  if (z < zmin) 
    lz = log(zmin);
  else
    lz = log(z);

  switch(data)
    {
    case BINARY_DATA:
       {
	double lz1;
	lz1 = EIGN[0] * lz;
	for(i = 0; i <  numberOfCategories; i++)
	  {		 
	    diagptable[2 * i] = 1.0;
	    diagptable[2 * i + 1] = exp(rptr[i] * lz1);	   	    
	  }
      }
      break;
    case DNA_DATA:
      {
	double lz1, lz2, lz3;
	lz1 = EIGN[0] * lz;
	lz2 = EIGN[1] * lz;
	lz3 = EIGN[2] * lz;

	for(i = 0; i <  numberOfCategories; i++)
	  {		 
	    diagptable[4 * i] = 1.0;
	    diagptable[4 * i + 1] = exp(rptr[i] * lz1);
	    diagptable[4 * i + 2] = exp(rptr[i] * lz2);
	    diagptable[4 * i + 3] = exp(rptr[i] * lz3);	    
	  }
      }
      break;
    case AA_DATA:
      {
	double lza[19];

	for(l = 0; l < 19; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 20] = 1.0;

	    for(l = 1; l < 20; l++)
	      diagptable[i * 20 + l] = exp(rptr[i] * lza[l - 1]);     	          
	  }
      }
      break;
    case SECONDARY_DATA:
      {
	double lza[15];

	for(l = 0; l < 15; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 16] = 1.0;

	    for(l = 1; l < 16; l++)
	      diagptable[i * 16 + l] = exp(rptr[i] * lza[l - 1]);     	          
	  }
      }
      break;
    case SECONDARY_DATA_6:
      {
	double lza[5];

	for(l = 0; l < 5; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 6] = 1.0;

	    for(l = 1; l < 6; l++)
	      diagptable[i * 6 + l] = exp(rptr[i] * lza[l - 1]);     	          
	  }
      }
      break;
    case SECONDARY_DATA_7:
      {
	double lza[6];

	for(l = 0; l < 6; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 7] = 1.0;

	    for(l = 1; l < 7; l++)
	      diagptable[i * 7 + l] = exp(rptr[i] * lza[l - 1]);     	          
	  }
      }
      break;
    default:
      assert(0);
    }
}


static void calcDiagptable_FLOAT(double z, int data, int numberOfCategories, double *rptr, double *EIGN, float *diagptable)
{
  int i, l;
  double lz;

  if (z < zmin) 
    lz = log(zmin);
  else
    lz = log(z);

  switch(data)
    {
    case DNA_DATA:         
      {
	double lz1, lz2, lz3;
	lz1 = EIGN[0] * lz;
	lz2 = EIGN[1] * lz;
	lz3 = EIGN[2] * lz;

	for(i = 0; i <  numberOfCategories; i++)
	  {		 
	    diagptable[4 * i] = 1.0;
	    diagptable[4 * i + 1] = (float)(exp(rptr[i] * lz1));
	    diagptable[4 * i + 2] = (float)(exp(rptr[i] * lz2));
	    diagptable[4 * i + 3] = (float)(exp(rptr[i] * lz3));	   
	  }
      }
      break;
    case BINARY_DATA:
      {
	double lz1;
	lz1 = EIGN[0] * lz;
	for(i = 0; i <  numberOfCategories; i++)
	  {		 
	    diagptable[2 * i] = 1.0;
	    diagptable[2 * i + 1] = (float)(exp(rptr[i] * lz1));	   	    
	  }
      }
      break;   
    case AA_DATA:
      {
	double lza[19];

	for(l = 0; l < 19; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 20] = 1.0;

	    for(l = 1; l < 20; l++)
	      diagptable[i * 20 + l] = (float)(exp(rptr[i] * lza[l - 1]));     	          
	  }
      }
      break;
    case SECONDARY_DATA:
      {
	double lza[15];

	for(l = 0; l < 15; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 16] = 1.0;

	    for(l = 1; l < 16; l++)
	      diagptable[i * 16 + l] = (float)(exp(rptr[i] * lza[l - 1]));     	          
	  }
      }
      break;
    case SECONDARY_DATA_6:
      {
	double lza[5];

	for(l = 0; l < 5; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 6] = 1.0;

	    for(l = 1; l < 6; l++)
	      diagptable[i * 6 + l] = (float)(exp(rptr[i] * lza[l - 1]));     	          
	  }
      }
      break;
    case SECONDARY_DATA_7:
      {
	double lza[6];

	for(l = 0; l < 6; l++)      
	  lza[l] = EIGN[l] * lz; 

	for(i = 0; i <  numberOfCategories; i++)
	  {	      	       
	    diagptable[i * 7] = 1.0;

	    for(l = 1; l < 7; l++)
	      diagptable[i * 7 + l] = (float)(exp(rptr[i] * lza[l - 1]));     	          
	  }
      }
      break;
    default:
      assert(0);
    }
}







static double evaluateGTRCATPROT (int *ex1, int *ex2, int *cptr, int *wptr,
				  double *x1, double *x2, double *tipVector,
				  unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;
  double  *diagptable,  *left, *right;
  int     i, l;                           
  
  if(tipX1)
    {            
      if(writeVector)       
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[20 * tipX1[i]]);
	    right = &(x2[20 * i]);
	    
	    diagptable = &diagptable_start[20 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 20; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
	    
	    term = log(term) + (ex2[i] * log(minlikelihood));
	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }     	
      else
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[20 * tipX1[i]]);
	    right = &(x2[20 * i]);
	    
	    diagptable = &diagptable_start[20 * cptr[i]];	           	 
#ifdef __SIM_SSE3
	    __m128d tv = _mm_setzero_pd();	    
	      	    
	    for(l = 0; l < 20; l+=2)
	      {
		__m128d lv = _mm_load_pd(&left[l]);
		__m128d rv = _mm_load_pd(&right[l]);
		__m128d mul = _mm_mul_pd(lv, rv);
		__m128d dv = _mm_load_pd(&diagptable[l]);
		
		tv = _mm_add_pd(tv, _mm_mul_pd(mul, dv));		   
	      }		 		
	      
	      tv = _mm_hadd_pd(tv, tv);
	      _mm_storel_pd(&term, tv);
#else  
	    for(l = 0, term = 0.0; l < 20; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
#endif	    
	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    sum += wptr[i] * term;
	  }      
    }    
  else
    {
    
      for (i = 0; i < n; i++) 
	{		       	      	      
	  left  = &x1[20 * i];
	  right = &x2[20 * i];
	  
	  diagptable = &diagptable_start[20 * cptr[i]];	  	
#ifdef __SIM_SSE3
	    __m128d tv = _mm_setzero_pd();	    
	      	    
	    for(l = 0; l < 20; l+=2)
	      {
		__m128d lv = _mm_load_pd(&left[l]);
		__m128d rv = _mm_load_pd(&right[l]);
		__m128d mul = _mm_mul_pd(lv, rv);
		__m128d dv = _mm_load_pd(&diagptable[l]);
		
		tv = _mm_add_pd(tv, _mm_mul_pd(mul, dv));		   
	      }		 		
	      
	      tv = _mm_hadd_pd(tv, tv);
	      _mm_storel_pd(&term, tv);
#else  
	  for(l = 0, term = 0.0; l < 20; l++)
	    term += left[l] * right[l] * diagptable[l];	
#endif

	  if(fastScaling)
	    term = log(term);
	  else
	    term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  
	  sum += wptr[i] * term;      
	}
    }
             
  return  sum;         
} 

static double evaluateGTRCATSECONDARY (int *ex1, int *ex2, int *cptr, int *wptr,
				       double *x1, double *x2, double *tipVector,
				       unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;
  double  *diagptable,  *left, *right;
  int     i, l;                           
  
  if(tipX1)
    {            
      if(writeVector)	  
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[16 * tipX1[i]]);
	    right = &(x2[16 * i]);
	    
	    diagptable = &diagptable_start[16 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 16; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
	    
	    term = log(term) + (ex2[i] * log(minlikelihood));
	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[16 * tipX1[i]]);
	    right = &(x2[16 * i]);
	    
	    diagptable = &diagptable_start[16 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 16; l++)
	      term += left[l] * right[l] * diagptable[l];
	 	  	  
	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    sum += wptr[i] * term;
	  }      
    }    
  else
    {
    
      for (i = 0; i < n; i++) 
	{		       	      	      
	  left  = &x1[16 * i];
	  right = &x2[16 * i];
	  
	  diagptable = &diagptable_start[16 * cptr[i]];	  	
	  
	  for(l = 0, term = 0.0; l < 16; l++)
	    term += left[l] * right[l] * diagptable[l];	
	  
	  if(fastScaling)
	    term = log(term);
	  else
	    term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  
	  sum += wptr[i] * term;      
	}
    }
             
  return  sum;         
} 

static double evaluateGTRCATSECONDARY_6 (int *ex1, int *ex2, int *cptr, int *wptr,
				       double *x1, double *x2, double *tipVector,
				       unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;
  double  *diagptable,  *left, *right;
  int     i, l;                           
  
  if(tipX1)
    {            
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[6 * tipX1[i]]);
	    right = &(x2[6 * i]);
	    
	    diagptable = &diagptable_start[6 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 6; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
	    
	    term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[6 * tipX1[i]]);
	    right = &(x2[6 * i]);
	    
	    diagptable = &diagptable_start[6 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 6; l++)
	      term += left[l] * right[l] * diagptable[l];
	    
	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    sum += wptr[i] * term;
	  }      
    }    
  else
    {
    
      for (i = 0; i < n; i++) 
	{		       	      	      
	  left  = &x1[6 * i];
	  right = &x2[6 * i];
	  
	  diagptable = &diagptable_start[6 * cptr[i]];	  	
	  
	  for(l = 0, term = 0.0; l < 6; l++)
	    term += left[l] * right[l] * diagptable[l];	
	  
	   if(fastScaling)
	     term = log(term);
	   else
	     term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  
	  sum += wptr[i] * term;      
	}
    }
             
  return  sum;         
} 

static double evaluateGTRCATSECONDARY_7(int *ex1, int *ex2, int *cptr, int *wptr,
					double *x1, double *x2, double *tipVector,
					unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;
  double  *diagptable,  *left, *right;
  int     i, l;                           
  
  if(tipX1)
    {            
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[7 * tipX1[i]]);
	    right = &(x2[7 * i]);
	    
	    diagptable = &diagptable_start[7 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 7; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
	    
	    term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {	       	
	    left = &(tipVector[7 * tipX1[i]]);
	    right = &(x2[7 * i]);
	    
	    diagptable = &diagptable_start[7 * cptr[i]];	           	 
	    
	    for(l = 0, term = 0.0; l < 7; l++)
	      term += left[l] * right[l] * diagptable[l];	 	  	  
	    
	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));
	   
	    sum += wptr[i] * term;
	  }      
    }    
  else
    {
    
      for (i = 0; i < n; i++) 
	{		       	      	      
	  left  = &x1[7 * i];
	  right = &x2[7 * i];
	  
	  diagptable = &diagptable_start[7 * cptr[i]];	  	
	  
	  for(l = 0, term = 0.0; l < 7; l++)
	    term += left[l] * right[l] * diagptable[l];	

	  if(fastScaling)
	    term = log(term);
	  else
	    term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  
	  sum += wptr[i] * term;      
	}
    }
             
  return  sum;         
} 

static double evaluateGTRCAT_BINARY (int *ex1, int *ex2, int *cptr, int *wptr,
				     double *x1_start, double *x2_start, double *tipVector, 		      
				     unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double  sum = 0.0, term;       
  int     i, j;  
  double  *diagptable, *x1, *x2;                      	    
 
  if(tipX1)
    {     
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {	    		   	  
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &x2_start[2 * i];
	    
	    diagptable = &diagptable_start[2 * cptr[i]];	    	    	  
	    
	    for(j = 0, term = 0.0; j < 2; j++)
	      term += x1[j] * x2[j] * diagptable[j];
	     
	    term = log(term) + (ex2[i] * log(minlikelihood));	   
	    
	    vector[i] = term;	   
	    
	    sum += wptr[i] * term;
	  }	
      else
	for (i = 0; i < n; i++) 
	  {	    		   	  
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &x2_start[2 * i];
	    
	    diagptable = &diagptable_start[2 * cptr[i]];	    	    	  
	    
	    for(j = 0, term = 0.0; j < 2; j++)       
	      term += x1[j] * x2[j] * diagptable[j];	      
	    
	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));	   	    	   	 	  	  	 
	    
	    sum += wptr[i] * term;
	  }	
    }               
  else
    {
      for (i = 0; i < n; i++) 
	{		          	
	  x1 = &x1_start[2 * i];
	  x2 = &x2_start[2 * i];
	  
	  diagptable = &diagptable_start[2 * cptr[i]];		  
	  
	  for(j = 0, term = 0.0; j < 2; j++)
	    term += x1[j] * x2[j] * diagptable[j];   
	  
	  if(fastScaling)
	    term = log(term);
	  else
	    term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  
	  sum += wptr[i] * term;
	}	   
    }
       
  return  sum;         
} 


static double evaluateGTRGAMMA_BINARY(int *ex1, int *ex2, int *wptr,
				      double *x1_start, double *x2_start, 
				      double *tipVector, 
				      unsigned char *tipX1, const int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;    
  int     i, j, k;
  double  *x1, *x2;             

  /* C-OPT once again switch to figure out if the node to the left or right of the branch at which we compute
     the final likelihood is a tip or not */

  if(tipX1)
    {    
      if(writeVector)      
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);	 
	    x2 = &x2_start[8 * i];	          	  	
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 2; k++)
		term += x1[k] * x2[j * 2 + k] * diagptable[j * 2 + k];	          	  	  
	    
	    term = log(0.25 * term) + ex2[i] * log(minlikelihood);	 
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }	  
      else	
	/* C-OPT loop over the input data length, this is where the interesting things happen */
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);	 
	    x2 = &x2_start[8 * i];	          	  	
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 2; k++)
		term += x1[k] * x2[j * 2 + k] * diagptable[j * 2 + k];	          	  	  	    	    

	    if(fastScaling)
	      term = log(0.25 * term);
	    else
	      term = log(0.25 * term) + ex2[i] * log(minlikelihood);	 

	    sum += wptr[i] * term;
	  }	  
    }
  else
    {   
      /* C-OPT loop over the input data length, this is where the interesting things happen */
      for (i = 0; i < n; i++) 
	{	  	 	  	  
	  x1 = &x1_start[8 * i];
	  x2 = &x2_start[8 * i];	  	  
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    for(k = 0; k < 2; k++)
	      term += x1[j * 2 + k] * x2[j * 2 + k] * diagptable[j * 2 + k];	          	  	  	      

	  if(fastScaling)
	    term = log(0.25 * term);
	  else
	    term = log(0.25 * term) + (ex1[i] + ex2[i]) * log(minlikelihood);

	  sum += wptr[i] * term;
	}                      	
    }

  return sum;
} 

static double evaluateGTRGAMMAINVAR_BINARY (int *ex1, int *ex2, int *wptr, int *iptr,
					    double *x1_start, double *x2_start,
					    double *tipVector, double *tFreqs, double invariants,
					    unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{ 
  int     i, j, k;
  double  *x1, *x2; 
  double 
    freqs[2], 
    scaler = 0.25 * (1.0 - invariants),
    sum = 0.0, 
    term; 

  freqs[0] = tFreqs[0] * invariants; 
  freqs[1] = tFreqs[1] * invariants; 

  if(tipX1)
    {   
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &x2_start[8 * i];	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 2; k++)
		term += x1[k] * x2[j * 2 + k] * diagptable[j * 2 + k];
	    
	    if(iptr[i] < 2)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));	 
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }	  
      else
	for (i = 0; i < n; i++) 
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &x2_start[8 * i];	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 2; k++)
		term += x1[k] * x2[j * 2 + k] * diagptable[j * 2 + k];
	    
	    if(iptr[i] < 2)
	      if(fastScaling)	   
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));	 
	    	   
	    sum += wptr[i] * term;
	  }	  
    }
  else
    {           		

      for (i = 0; i < n; i++) 
	{	  	 	  	
	  x1 = &x1_start[8 * i];
	  x2 = &x2_start[8 * i];	  
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    for(k = 0; k < 2; k++)
	      term += x1[j * 2 + k] * x2[j * 2 + k] * diagptable[j * 2 + k];	  	 	      
	  
	  if(iptr[i] < 2)
	    if(fastScaling)	   	       
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]])) + (ex2[i] + ex1[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + ((ex1[i] + ex2[i]) * log(minlikelihood));

	  sum += wptr[i] * term;
	}	  	                        
    }

  return  sum;
} 


static double evaluateGTRCAT (int *ex1, int *ex2, int *cptr, int *wptr,
			      double *x1_start, double *x2_start, double *tipVector, 		      
			      unsigned char *tipX1, int n, double *diagptable_start, double *vector, boolean writeVector, const boolean fastScaling)
{
  double  sum = 0.0, term;       
  int     i, j;  
  double  *diagptable, *x1, *x2;                      	    
 
  if(tipX1)
    {     
      if(writeVector)       
	for (i = 0; i < n; i++) 
	  {
#ifdef __SIM_SSE3
	    double t[2] __attribute__ ((aligned (16)));
	    __m128d x1v1, x1v2, x2v1, x2v2, dv1, dv2;
#endif	    		   	  
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    
	    diagptable = &diagptable_start[4 * cptr[i]];	    	    	  
	    
#ifdef __SIM_SSE3	    	  
	    x1v1 =  _mm_load_pd(&x1[0]);
	    x1v2 =  _mm_load_pd(&x1[2]);
	    x2v1 =  _mm_load_pd(&x2[0]);
	    x2v2 =  _mm_load_pd(&x2[2]);
	    dv1  =  _mm_load_pd(&diagptable[0]);
	    dv2  =  _mm_load_pd(&diagptable[2]);
	    
	    x1v1 = _mm_mul_pd(x1v1, x2v1);
	    x1v1 = _mm_mul_pd(x1v1, dv1);
	    
	    x1v2 = _mm_mul_pd(x1v2, x2v2);
	    x1v2 = _mm_mul_pd(x1v2, dv2);
	    
	    x1v1 = _mm_add_pd(x1v1, x1v2);
	    
	    _mm_store_pd(t, x1v1);
	    
	    term = log(t[0] + t[1]) + (ex2[i] * log(minlikelihood));
#else
	    for(j = 0, term = 0.0; j < 4; j++)
	      term += x1[j] * x2[j] * diagptable[j];
	    
	    term = log(term) + (ex2[i] * log(minlikelihood));	   
#endif
	    
	    vector[i] = term;	   
	    
	    sum += wptr[i] * term;
	  }	    
      else
	for (i = 0; i < n; i++) 
	  {	
#ifdef __SIM_SSE3
	    double t[2] __attribute__ ((aligned (16)));
	    __m128d x1v1, x1v2, x2v1, x2v2, dv1, dv2;
#endif
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    
	    diagptable = &diagptable_start[4 * cptr[i]];
	    	    	  
#ifdef __SIM_SSE3	    	  
	    x1v1 =  _mm_load_pd(&x1[0]);
	    x1v2 =  _mm_load_pd(&x1[2]);
	    x2v1 =  _mm_load_pd(&x2[0]);
	    x2v2 =  _mm_load_pd(&x2[2]);
	    dv1  =  _mm_load_pd(&diagptable[0]);
	    dv2  =  _mm_load_pd(&diagptable[2]);
	    
	    x1v1 = _mm_mul_pd(x1v1, x2v1);
	    x1v1 = _mm_mul_pd(x1v1, dv1);

	    x1v2 = _mm_mul_pd(x1v2, x2v2);
	    x1v2 = _mm_mul_pd(x1v2, dv2);

	    x1v1 = _mm_add_pd(x1v1, x1v2);

	    _mm_store_pd(t, x1v1);
	    	    	    
	    if(fastScaling)
	      term = log(t[0] + t[1]);
	    else
	      term = log(t[0] + t[1]) + (ex2[i] * log(minlikelihood));
#else
	    for(j = 0, term = 0.0; j < 4; j++)
	      term += x1[j] * x2[j] * diagptable[j];

	    if(fastScaling)
	      term = log(term);
	    else
	      term = log(term) + (ex2[i] * log(minlikelihood));
#endif
	    
	    sum += wptr[i] * term;
	  }	
    }               
  else
    {
      for (i = 0; i < n; i++) 
	{ 
#ifdef __SIM_SSE3
	  double t[2] __attribute__ ((aligned (16)));
	   __m128d x1v1, x1v2, x2v1, x2v2, dv1, dv2;
#endif
	  x1 = &x1_start[4 * i];
	  x2 = &x2_start[4 * i];
	  
	  diagptable = &diagptable_start[4 * cptr[i]];	
	  
#ifdef __SIM_SSE3	  
	  x1v1 =  _mm_load_pd(&x1[0]);
	  x1v2 =  _mm_load_pd(&x1[2]);
	  x2v1 =  _mm_load_pd(&x2[0]);
	  x2v2 =  _mm_load_pd(&x2[2]);
	  dv1  =  _mm_load_pd(&diagptable[0]);
	  dv2  =  _mm_load_pd(&diagptable[2]);
	  
	  x1v1 = _mm_mul_pd(x1v1, x2v1);
	  x1v1 = _mm_mul_pd(x1v1, dv1);
	  
	  x1v2 = _mm_mul_pd(x1v2, x2v2);
	  x1v2 = _mm_mul_pd(x1v2, dv2);
	  
	  x1v1 = _mm_add_pd(x1v1, x1v2);
	  
	  _mm_store_pd(t, x1v1);
	  
	  if(fastScaling)
	    term = log(t[0] + t[1]);
	  else
	    term = log(t[0] + t[1]) + ((ex1[i] + ex2[i]) * log(minlikelihood));
#else
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    term += x1[j] * x2[j] * diagptable[j];     
	  
	  if(fastScaling)
	    term = log(term);
	  else
	    term = log(term) + ((ex1[i] + ex2[i]) * log(minlikelihood));	  
#endif
	  sum += wptr[i] * term;
	}    
    }
       
  return  sum;         
} 

static double evaluateGTRCAT_FLOAT (int *ex1, int *ex2, int *cptr, int *wptr,
				    float *x1_start, float *x2_start, float *tipVector, 		      
				    unsigned char *tipX1, int n, float *diagptable_start, float *vector, boolean writeVector, const boolean fastScaling)
{
  float  sum = 0.0, term;       
  int     i, j;  
  float  *diagptable, *x1, *x2;                      	    
 
  if(tipX1)
    {     
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {	    		   	  
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    
	    diagptable = &diagptable_start[4 * cptr[i]];	    	    	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      term += x1[j] * x2[j] * diagptable[j];
	     
	    term = logf(term) + (ex2[i] * logf(minlikelihood_FLOAT));	   
	    
	    vector[i] = term;	   
	    
	    sum += wptr[i] * term;
	  }	
      else
	for (i = 0; i < n; i++) 
	  {		   		   
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    
	    diagptable = &diagptable_start[4 * cptr[i]];	    	    	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      term += x1[j] * x2[j] * diagptable[j];

	    if(fastScaling)	   	       
	      term = logf(term);
	    else
	      term = logf(term) + (ex2[i] * logf(minlikelihood_FLOAT));	   	    	   	 	  	  	 

	    sum += wptr[i] * term;
	  }	
    }               
  else
    {
      for (i = 0; i < n; i++) 
	{	 	           	
	  x1 = &x1_start[4 * i];
	  x2 = &x2_start[4 * i];
	  
	  diagptable = &diagptable_start[4 * cptr[i]];		  
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    term += x1[j] * x2[j] * diagptable[j];     	
	  
	  if(fastScaling)	   	       
	    term = logf(term);
	  else
	    term = logf(term) + ((ex1[i] + ex2[i]) * logf(minlikelihood_FLOAT));	  

	  sum += wptr[i] * term;
	}    
    }        

  return  ((double)sum);         
} 


static double evaluateGTRGAMMA(int *ex1, int *ex2, int *wptr,
			       double *x1_start, double *x2_start, 
			       double *tipVector, 
			       unsigned char *tipX1, const int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;    
  int     i, j, k;
  double  *x1, *x2;             

 

  if(tipX1)
    {    
      if(writeVector)      
	for (i = 0; i < n; i++)
	  {
#ifdef __SIM_SSE3
	    double t[2] __attribute__ ((aligned (16)));
	    __m128d termv, x1v, x2v, dv;
#endif
	    x1 = &(tipVector[4 * tipX1[i]]);	 
	    x2 = &x2_start[16 * i];
	          	  	
#ifdef __SIM_SSE3	
	    termv = _mm_set1_pd(0.0);	    
    
	    for(j = 0; j < 4; j++)
	      {
		x1v = _mm_load_pd(&x1[0]);
		x2v = _mm_load_pd(&x2[j * 4]);
		dv   = _mm_load_pd(&diagptable[j * 4]);

		x1v = _mm_mul_pd(x1v, x2v);
		x1v = _mm_mul_pd(x1v, dv);

		termv = _mm_add_pd(termv, x1v);

		x1v = _mm_load_pd(&x1[2]);
		x2v = _mm_load_pd(&x2[j * 4 + 2]);
		dv   = _mm_load_pd(&diagptable[j * 4 + 2]);

		x1v = _mm_mul_pd(x1v, x2v);
		x1v = _mm_mul_pd(x1v, dv);

		termv = _mm_add_pd(termv, x1v);
	      }

	    _mm_store_pd(t, termv);
			    
	    term = log(0.25 * (t[0] + t[1])) + (ex2[i] * log(minlikelihood));   
#else	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];	          	  	  

	    term = log(0.25 * term) + ex2[i] * log(minlikelihood);	 
#endif
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }	  
      else		
	for (i = 0; i < n; i++)
	  {
#ifdef __SIM_SSE3
	    double t[2] __attribute__ ((aligned (16)));
	    __m128d termv, x1v, x2v, dv;
#endif
	    x1 = &(tipVector[4 * tipX1[i]]);	 
	    x2 = &x2_start[16 * i];	 

#ifdef __SIM_SSE3	
	    termv = _mm_set1_pd(0.0);	    	   
    
	    for(j = 0; j < 4; j++)
	      {
		x1v = _mm_load_pd(&x1[0]);
		x2v = _mm_load_pd(&x2[j * 4]);
		dv   = _mm_load_pd(&diagptable[j * 4]);

		x1v = _mm_mul_pd(x1v, x2v);
		x1v = _mm_mul_pd(x1v, dv);

		termv = _mm_add_pd(termv, x1v);

		x1v = _mm_load_pd(&x1[2]);
		x2v = _mm_load_pd(&x2[j * 4 + 2]);
		dv   = _mm_load_pd(&diagptable[j * 4 + 2]);

		x1v = _mm_mul_pd(x1v, x2v);
		x1v = _mm_mul_pd(x1v, dv);

		termv = _mm_add_pd(termv, x1v);
	      }

	    _mm_store_pd(t, termv);
		

	    if(fastScaling)
	      term = log(0.25 * (t[0] + t[1]));
	    else
	      term = log(0.25 * (t[0] + t[1])) + (ex2[i] * log(minlikelihood));
    
#else
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];	          	  	  	    	    	  

	    if(fastScaling)
	      term = log(0.25 * term);
	    else
	      term = log(0.25 * term) + ex2[i] * log(minlikelihood);	 
#endif

	    sum += wptr[i] * term;
	  }     
    }
  else
    {        
      for (i = 0; i < n; i++) 
	{
#ifdef __SIM_SSE3
	  double t[2] __attribute__ ((aligned (16)));
	  __m128d termv, x1v, x2v, dv;
#endif
	  	 	  	  
	  x1 = &x1_start[16 * i];
	  x2 = &x2_start[16 * i];	  	  
	
#ifdef __SIM_SSE3	
	  termv = _mm_set1_pd(0.0);	  	 
	  
	  for(j = 0; j < 4; j++)
	    {
	      x1v = _mm_load_pd(&x1[j * 4]);
	      x2v = _mm_load_pd(&x2[j * 4]);
	      dv   = _mm_load_pd(&diagptable[j * 4]);
	      
	      x1v = _mm_mul_pd(x1v, x2v);
	      x1v = _mm_mul_pd(x1v, dv);
	      
	      termv = _mm_add_pd(termv, x1v);
	      
	      x1v = _mm_load_pd(&x1[j * 4 + 2]);
	      x2v = _mm_load_pd(&x2[j * 4 + 2]);
	      dv   = _mm_load_pd(&diagptable[j * 4 + 2]);
	      
	      x1v = _mm_mul_pd(x1v, x2v);
	      x1v = _mm_mul_pd(x1v, dv);
	      
	      termv = _mm_add_pd(termv, x1v);
	    }
	  
	  _mm_store_pd(t, termv);

	  if(fastScaling)
	    term = log(0.25 * (t[0] + t[1]));
	  else
	    term = log(0.25 * (t[0] + t[1])) + ((ex1[i] + ex2[i]) * log(minlikelihood));	  
#else 
	  for(j = 0, term = 0.0; j < 4; j++)
	    for(k = 0; k < 4; k++)
	      term += x1[j * 4 + k] * x2[j * 4 + k] * diagptable[j * 4 + k];
	          	  	  	      
	   if(fastScaling)
	      term = log(0.25 * term);
	    else
	      term = log(0.25 * term) + (ex1[i] + ex2[i]) * log(minlikelihood);
#endif
	  
	  sum += wptr[i] * term;
	}                      	
    }

  return sum;
} 

static double evaluateGTRGAMMA_FLOAT(int *ex1, int *ex2, int *wptr,
				     float *x1_start, float *x2_start, 
				     float *tipVector, 
				     unsigned char *tipX1, const int n, float *diagptable, float *vector, boolean writeVector, const boolean fastScaling)
{
  float   sum = 0.0, term;    
  int     i, j, k;
  float  *x1, *x2;             

  /* C-OPT once again switch to figure out if the node to the left or right of the branch at which we compute
     the final likelihood is a tip or not */

  if(tipX1)
    {    
      if(writeVector)      
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);	 
	    x2 = &x2_start[16 * i];	          	  	
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];	          	  	  
	    
	    term = logf(0.25 * term) + ex2[i] * logf(minlikelihood_FLOAT);	 
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }	  
      else	
	/* C-OPT loop over the input data length, this is where the interesting things happen */
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);	 
	    x2 = &x2_start[16 * i];	          	  	
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];	          	  	  	    	    	    

	    if(fastScaling)
	      term = logf(0.25 * term);
	    else
	      term = logf(0.25 * term) + ex2[i] * logf(minlikelihood_FLOAT);	 

	    sum += wptr[i] * term;
	  }     
    }
  else
    {         
      for (i = 0; i < n; i++) 
	{	  	 	  	  
	  x1 = &x1_start[16 * i];
	  x2 = &x2_start[16 * i];	  	  
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    for(k = 0; k < 4; k++)
	      term += x1[j * 4 + k] * x2[j * 4 + k] * diagptable[j * 4 + k];
	  
	  if(fastScaling)
	    term = logf(0.25 * term);
	  else
	    term = logf(0.25 * term) + (ex1[i] + ex2[i]) * logf(minlikelihood_FLOAT);

	  sum += wptr[i] * term;
	}                      	
    }

  return ((double)sum);
} 






static double evaluateGTRGAMMAINVAR (int *ex1, int *ex2, int *wptr, int *iptr,
				     double *x1_start, double *x2_start,
				     double *tipVector, double *tFreqs, double invariants,
				     unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{ 
  int     i, j, k;
  double  *x1, *x2; 
  double 
    freqs[4], 
    scaler = 0.25 * (1.0 - invariants),
    sum = 0.0, 
    term; 

  freqs[0] = tFreqs[0] * invariants; 
  freqs[1] = tFreqs[1] * invariants;
  freqs[2] = tFreqs[2] * invariants;
  freqs[3] = tFreqs[3] * invariants;   

  if(tipX1)
    {   
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[16 * i];	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];
	    
	    if(iptr[i] < 4)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));	 
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }	  
      else
	for (i = 0; i < n; i++) 
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[16 * i];	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		term += x1[k] * x2[j * 4 + k] * diagptable[j * 4 + k];
	    
	    if(iptr[i] < 4)
	      if(fastScaling)
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));	 
	    	   
	    sum += wptr[i] * term;
	  }	  
    }
  else
    {           		

      for (i = 0; i < n; i++) 
	{	  	 	  	
	  x1 = &x1_start[16 * i];
	  x2 = &x2_start[16 * i];	  
	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    for(k = 0; k < 4; k++)
	      term += x1[j * 4 + k] * x2[j * 4 + k] * diagptable[j * 4 + k];	  	 	      
	  
	  if(iptr[i] < 4)
	    if(fastScaling)
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]])) + (ex2[i] + ex1[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + ((ex1[i] + ex2[i]) * log(minlikelihood));

	  sum += wptr[i] * term;
	}	  	                        
    }

  return  sum;
} 




static double evaluateGTRGAMMAPROT (int *ex1, int *ex2, int *wptr,
				    double *x1, double *x2,  
				    double *tipVector, 
				    unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;        
  int     i, j, l;   
  double  *left, *right;              
  
  if(tipX1)
    {          
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[20 * tipX1[i]]);	  	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[80 * i + 20 * j]);
		
	      for(l = 0; l < 20; l++)
		term += left[l] * right[l] * diagptable[j * 20 + l];	      
	      }	  
	    
	    term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	{       
	  for (i = 0; i < n; i++) 
	    {
#ifdef __SIM_SSE3
	      __m128d tv = _mm_setzero_pd();
	      left = &(tipVector[20 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		  double *d = &diagptable[j * 20];
		  right = &(x2[80 * i + 20 * j]);
		  for(l = 0; l < 20; l+=2)
		    {
		      __m128d mul = _mm_mul_pd(_mm_load_pd(&left[l]), _mm_load_pd(&right[l]));
		      tv = _mm_add_pd(tv, _mm_mul_pd(mul, _mm_load_pd(&d[l])));		   
		    }		 		
		}
	      tv = _mm_hadd_pd(tv, tv);
	      _mm_storel_pd(&term, tv);

#else
	      left = &(tipVector[20 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		 right = &(x2[80 * i + 20 * j]);
		  for(l = 0; l < 20; l++)
		    term += left[l] * right[l] * diagptable[j * 20 + l];	      
		}	  
#endif
	      
	      if(fastScaling)
		term = log(0.25 * term);
	      else
		term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	      
	      sum += wptr[i] * term;
	    }     	 
	}
    }              
  else
    {
      for (i = 0; i < n; i++) 
	{	  	 	             
#ifdef __SIM_SSE3
	  __m128d tv = _mm_setzero_pd();	 	  	  
	      
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      double *d = &diagptable[j * 20];
	      left  = &(x1[80 * i + 20 * j]);
	      right = &(x2[80 * i + 20 * j]);
	      
	      for(l = 0; l < 20; l+=2)
		{
		  __m128d mul = _mm_mul_pd(_mm_load_pd(&left[l]), _mm_load_pd(&right[l]));
		  tv = _mm_add_pd(tv, _mm_mul_pd(mul, _mm_load_pd(&d[l])));		   
		}		 		
	    }
	  tv = _mm_hadd_pd(tv, tv);
	  _mm_storel_pd(&term, tv);	  
#else
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[80 * i + 20 * j]);
	      right = &(x2[80 * i + 20 * j]);	    
	      
	      for(l = 0; l < 20; l++)
		term += left[l] * right[l] * diagptable[j * 20 + l];	
	    }
#endif
	  
	  if(fastScaling)
	    term = log(0.25 * term);
	  else
	    term = log(0.25 * term) + ((ex1[i] + ex2[i])*log(minlikelihood));
	  
	  sum += wptr[i] * term;
	}         
    }
       
  return  sum;
}

static double evaluateGTRGAMMAPROT_FLOAT (int *ex1, int *ex2, int *wptr,
					  float *x1, float *x2,  
					  float *tipVector, 
					  unsigned char *tipX1, int n, float *diagptable, float *vector, boolean writeVector, const boolean fastScaling)
{
  float   
    sum = 0.0, 
    term,
    *left,
    *right;
    
  int i, j, l;   
   
  if(tipX1)
    {          
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[20 * tipX1[i]]);	  	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[80 * i + 20 * j]);
		
	      for(l = 0; l < 20; l++)
		term += left[l] * right[l] * diagptable[j * 20 + l];	      
	      }	  
	    
	    term = logf(0.25 * term) + (ex2[i] * logf(minlikelihood_FLOAT));	   
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	{       
	  for (i = 0; i < n; i++) 
	    {	     
	      left = &(tipVector[20 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		  right = &(x2[80 * i + 20 * j]);
		  
		  for(l = 0; l < 20; l++)
		    term += left[l] * right[l] * diagptable[j * 20 + l];	      
		}	  
	      
	      if(fastScaling)
		term = logf(0.25 * term);
	      else
		term = logf(0.25 * term) + (ex2[i] * logf(minlikelihood_FLOAT));	   
	      
	      sum += wptr[i] * term;
	    }     	 
	}
    }              
  else
    {
      for (i = 0; i < n; i++) 
	{	  	 	             
      
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[80 * i + 20 * j]);
	      right = &(x2[80 * i + 20 * j]);	    
	      
	      for(l = 0; l < 20; l++)
		term += left[l] * right[l] * diagptable[j * 20 + l];	
	    }
	  
	  if(fastScaling)
	    term = logf(0.25 * term);
	  else
	    term = logf(0.25 * term) + ((ex1[i] + ex2[i])*logf(minlikelihood_FLOAT));
	  
	  sum += wptr[i] * term;
	}         
    }
       
  return  ((double)sum);
}



static double evaluateGTRGAMMASECONDARY (int *ex1, int *ex2, int *wptr,
					 double *x1, double *x2,  
					 double *tipVector, 
					 unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;        
  int     i, j, l;   
  double  *left, *right;              
  
  if(tipX1)
    {          
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[16 * tipX1[i]]);	  	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[64 * i + 16 * j]);
		
		for(l = 0; l < 16; l++)
		  term += left[l] * right[l] * diagptable[j * 16 + l];	      
	      }	  
	    
	    term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	{       
	  for (i = 0; i < n; i++) 
	    {	     
	      left = &(tipVector[16 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		  right = &(x2[64 * i + 16 * j]);
		  
		  for(l = 0; l < 16; l++)
		    term += left[l] * right[l] * diagptable[j * 16 + l];	      
		}
	      
	      if(fastScaling)
		term = log(0.25 * term);
	      else
		term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	      
	      sum += wptr[i] * term;
	    }     	 
	}
    }              
  else
    {
      for (i = 0; i < n; i++) 
	{	  	 	             
      
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[64 * i + 16 * j]);
	      right = &(x2[64 * i + 16 * j]);	    
	      
	      for(l = 0; l < 16; l++)
		term += left[l] * right[l] * diagptable[j * 16 + l];	
	    }
	  
	  if(fastScaling)
	    term = log(0.25 * term);
	  else
	    term = log(0.25 * term) + ((ex1[i] + ex2[i])*log(minlikelihood));
	  
	  sum += wptr[i] * term;
	}         
    }
       
  return  sum;
}

static double evaluateGTRGAMMASECONDARY_6 (int *ex1, int *ex2, int *wptr,
					   double *x1, double *x2,  
					   double *tipVector, 
					   unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;        
  int     i, j, l;   
  double  *left, *right;              
  
  if(tipX1)
    {          
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[6 * tipX1[i]]);	  	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[24 * i + 6 * j]);
		
		for(l = 0; l < 6; l++)
		  term += left[l] * right[l] * diagptable[j * 6 + l];	      
	      }	  
	    
	    term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	{       
	  for (i = 0; i < n; i++) 
	    {	     
	      left = &(tipVector[6 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		  right = &(x2[24 * i + 6 * j]);
		  
		  for(l = 0; l < 6; l++)
		    term += left[l] * right[l] * diagptable[j * 6 + l];	      
		}	
	      
	      if(fastScaling)
		term = log(0.25 * term);
	      else
		term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	      
	      sum += wptr[i] * term;
	    }     	 
	}
    }              
  else
    {
      for (i = 0; i < n; i++) 
	{	  	 	             
      
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[24 * i + 6 * j]);
	      right = &(x2[24 * i + 6 * j]);	    
	      
	      for(l = 0; l < 6; l++)
		term += left[l] * right[l] * diagptable[j * 6 + l];	
	    }

	  if(fastScaling)
	    term = log(0.25 * term);
	  else
	    term = log(0.25 * term) + ((ex1[i] + ex2[i])*log(minlikelihood));
	  
	  sum += wptr[i] * term;
	}         
    }
       
  return  sum;
}

static double evaluateGTRGAMMASECONDARY_7 (int *ex1, int *ex2, int *wptr,
					   double *x1, double *x2,  
					   double *tipVector, 
					   unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   sum = 0.0, term;        
  int     i, j, l;   
  double  *left, *right;              
  
  if(tipX1)
    {          
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[7 * tipX1[i]]);	  	  
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[28 * i + 7 * j]);
		
		for(l = 0; l < 7; l++)
		  term += left[l] * right[l] * diagptable[j * 7 + l];	      
	      }	  
	    
	    term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	    	    
	    vector[i] = term;
	    
	    sum += wptr[i] * term;
	  }         
      else
	{       
	  for (i = 0; i < n; i++) 
	    {	     
	      left = &(tipVector[7 * tipX1[i]]);	  	  
	      
	      for(j = 0, term = 0.0; j < 4; j++)
		{
		  right = &(x2[28 * i + 7 * j]);
		  
		  for(l = 0; l < 7; l++)
		    term += left[l] * right[l] * diagptable[j * 7 + l];	      
		}	
	      
	      if(fastScaling)
		term = log(0.25 * term);
	      else
		term = log(0.25 * term) + (ex2[i] * log(minlikelihood));	   
	      
	      sum += wptr[i] * term;
	    }     	 
	}
    }              
  else
    {
      for (i = 0; i < n; i++) 
	{	  	 	             
      
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[28 * i + 7 * j]);
	      right = &(x2[28 * i + 7 * j]);	    
	      
	      for(l = 0; l < 7; l++)
		term += left[l] * right[l] * diagptable[j * 7 + l];	
	    }
	  
	  if(fastScaling)
	    term = log(0.25 * term);
	  else
	    term = log(0.25 * term) + ((ex1[i] + ex2[i])*log(minlikelihood));
	  
	  sum += wptr[i] * term;
	}         
    }
       
  return  sum;
}

static double evaluateGTRGAMMAPROTINVAR (int *ex1, int *ex2, int *wptr, int *iptr,
					 double *x1, double *x2, 
					 double *tipVector,double *tFreqs, double invariants,
					 unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   
    sum = 0.0, term, freqs[20],
    scaler = 0.25 * (1.0 - invariants);        
  int     i, j, l;     
  double *left, *right;   
    
  for(i = 0; i < 20; i++)
    freqs[i] = tFreqs[i] * invariants;            	  
  
  if(tipX1)
    {    
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[20 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[80 * i + 20 * j]);
		
		for(l = 0; l < 20; l++)
		  term += left[l] * right[l] * diagptable[j * 20 + l];	      
	      }	  
	    
	    if(iptr[i] < 20)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    vector[i] = term;
	   
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[20 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[80 * i + 20 * j]);
		
		for(l = 0; l < 20; l++)
		  term += left[l] * right[l] * diagptable[j * 20 + l];	      
	      }	  
	    
	    if(iptr[i] < 20)	
	      if(fastScaling)
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]]))  + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    sum += wptr[i] * term;
	  }    	
    }                
  else
    {    
      for (i = 0; i < n; i++) 
	{	  	 	       	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[80 * i + 20 * j]);
	      right = &(x2[80 * i + 20 * j]);	    
	      
	      for(l = 0; l < 20; l++)
		term += left[l] * right[l] * diagptable[j * 20 + l];	
	    }
	  
	  if(iptr[i] < 20)
	    if(fastScaling)
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]])) + (ex1[i] + ex2[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + ((ex1[i] + ex2[i]) * log(minlikelihood));
	  sum += wptr[i] * term;
	}              
    }
       
  return  sum;
}

static double evaluateGTRGAMMASECONDARYINVAR (int *ex1, int *ex2, int *wptr, int *iptr,
					      double *x1, double *x2, 
					      double *tipVector,double *tFreqs, double invariants,
					      unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   
    sum = 0.0, term, freqs[16],
    scaler = 0.25 * (1.0 - invariants);        
  int     i, j, l;     
  double *left, *right;   
    
  for(i = 0; i < 16; i++)
    freqs[i] = tFreqs[i] * invariants;            	  
  
  if(tipX1)
    {    
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[16 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[64 * i + 16 * j]);
		
		for(l = 0; l < 16; l++)
		  term += left[l] * right[l] * diagptable[j * 16 + l];	      
	      }	  
	    
	    if(iptr[i] < 16)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    vector[i] = term;
	   
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[16 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[64 * i + 16 * j]);
		
		for(l = 0; l < 16; l++)
		  term += left[l] * right[l] * diagptable[j * 16 + l];	      
	      }	  
	    
	    if(iptr[i] < 16)
	      if(fastScaling) 
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]]))  + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    sum += wptr[i] * term;
	  }    	
    }                
  else
    {    
      for (i = 0; i < n; i++) 
	{	  	 	       	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[64 * i + 16 * j]);
	      right = &(x2[64 * i + 16 * j]);	    
	      
	      for(l = 0; l < 16; l++)
		term += left[l] * right[l] * diagptable[j * 16 + l];	
	    }

	  if(iptr[i] < 16)
	    if(fastScaling) 
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]]))  + (ex1[i] + ex2[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + (ex1[i] + ex2[i]) * log(minlikelihood);	  	 	
	  
	  sum += wptr[i] * term;
	}              
    }
       
  return  sum;
}

static double evaluateGTRGAMMASECONDARYINVAR_6 (int *ex1, int *ex2, int *wptr, int *iptr,
						double *x1, double *x2, 
						double *tipVector,double *tFreqs, double invariants,
						unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   
    sum = 0.0, term, freqs[6],
    scaler = 0.25 * (1.0 - invariants);        
  int     i, j, l;     
  double *left, *right;   
    
  for(i = 0; i < 6; i++)
    freqs[i] = tFreqs[i] * invariants;            	  
  
  if(tipX1)
    {    
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[6 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[24 * i + 6 * j]);
		
		for(l = 0; l < 6; l++)
		  term += left[l] * right[l] * diagptable[j * 6 + l];	      
	      }	  
	    
	    if(iptr[i] < 6)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    vector[i] = term;
	   
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[6 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[24 * i + 6 * j]);
		
		for(l = 0; l < 6; l++)
		  term += left[l] * right[l] * diagptable[j * 6 + l];	      
	      }	  
	    
	    if(iptr[i] < 6)
	      if(fastScaling)
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]]))  + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    sum += wptr[i] * term;
	  }    	
    }                
  else
    {    
      for (i = 0; i < n; i++) 
	{	  	 	       	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[24 * i + 6 * j]);
	      right = &(x2[24 * i + 6 * j]);	    
	      
	      for(l = 0; l < 6; l++)
		term += left[l] * right[l] * diagptable[j * 6 + l];	
	    }
	  
	  if(iptr[i] < 6)
	    if(fastScaling)
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]])) + (ex2[i] + ex1[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + ((ex1[i] + ex2[i]) * log(minlikelihood));

	  sum += wptr[i] * term;
	}              
    }
       
  return  sum;
}

static double evaluateGTRGAMMASECONDARYINVAR_7 (int *ex1, int *ex2, int *wptr, int *iptr,
						double *x1, double *x2, 
						double *tipVector,double *tFreqs, double invariants,
						unsigned char *tipX1, int n, double *diagptable, double *vector, boolean writeVector, const boolean fastScaling)
{
  double   
    sum = 0.0, term, freqs[7],
    scaler = 0.25 * (1.0 - invariants);        
  int     i, j, l;     
  double *left, *right;   
    
  for(i = 0; i < 7; i++)
    freqs[i] = tFreqs[i] * invariants;            	  
  
  if(tipX1)
    {    
      if(writeVector)
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[7 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[28 * i + 7 * j]);
		
		for(l = 0; l < 7; l++)
		  term += left[l] * right[l] * diagptable[j * 7 + l];	      
	      }	  
	    
	    if(iptr[i] < 7)	   
	      term = log(((scaler * term) + freqs[iptr[i]])) + ex2[i] * log(minlikelihood);
	    else
	      term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    vector[i] = term;
	   
	    sum += wptr[i] * term;
	  }         
      else
	for (i = 0; i < n; i++) 
	  {
	    left = &(tipVector[7 * tipX1[i]]);
	    
	    for(j = 0, term = 0.0; j < 4; j++)
	      {
		right = &(x2[28 * i + 7 * j]);
		
		for(l = 0; l < 7; l++)
		  term += left[l] * right[l] * diagptable[j * 7 + l];	      
	      }	  
	    
	    if(iptr[i] < 7)
	      if(fastScaling)
		term = log(((scaler * term) + freqs[iptr[i]]));
	      else
		term = log(((scaler * term) + freqs[iptr[i]]))  + ex2[i] * log(minlikelihood);
	    else
	      if(fastScaling)
		term = log(scaler * term);
	      else
		term = log(scaler * term) + (ex2[i] * log(minlikelihood));
	    	    
	    sum += wptr[i] * term;
	  }    	
    }                
  else
    {    
      for (i = 0; i < n; i++) 
	{	  	 	       	  
	  for(j = 0, term = 0.0; j < 4; j++)
	    {
	      left  = &(x1[28 * i + 7 * j]);
	      right = &(x2[28 * i + 7 * j]);	    
	      
	      for(l = 0; l < 7; l++)
		term += left[l] * right[l] * diagptable[j * 7 + l];	
	    }
	  
	  if(iptr[i] < 7)
	    if(fastScaling)
	      term = log(((scaler * term) + freqs[iptr[i]]));
	    else
	      term = log(((scaler * term) + freqs[iptr[i]])) + (ex2[i] + ex1[i]) * log(minlikelihood);
	  else
	    if(fastScaling)
	      term = log(scaler * term);
	    else
	      term = log(scaler * term) + ((ex1[i] + ex2[i]) * log(minlikelihood));

	  sum += wptr[i] * term;
	}              
    }
       
  return  sum;
}


double evaluateIterative(tree *tr,  boolean writeVector)
{
  double 
    result = 0.0;  
  int pNumber, qNumber, model;
  double *pz;

  pNumber = tr->td[0].ti[0].pNumber;
  qNumber = tr->td[0].ti[0].qNumber;
  pz      = tr->td[0].ti[0].qz;

  newviewIterative(tr); 

  if(writeVector)
    assert(!tr->useFastScaling);

  for(model = 0; model < tr->NumberOfModels; model++)
    {            
      if(tr->executeModel[model])
	{	
	  int width = tr->partitionData[model].width;
	  double z, partitionLikelihood, *_vector;
	  int    
	    *ex1 = (int*)NULL, 
	    *ex2 = (int*)NULL;

	  double 
	    *x1_start   = (double*)NULL, 
	    *x2_start   = (double*)NULL,
	    *diagptable = (double*)NULL;

	  float 
	    *x1_start_FLOAT   = (float*)NULL, 
	    *x2_start_FLOAT   = (float*)NULL,
	    *_vector_FLOAT    = (float*)NULL,
	    *diagptable_FLOAT = (float*)NULL;


	  unsigned char 
	    *tip = (unsigned char*)NULL;

	  if(writeVector)
	    _vector = tr->partitionData[model].perSiteLL;
	  else
	    _vector = (double*)NULL;

	  if(tr->useFloat)
	    diagptable_FLOAT = tr->partitionData[model].left_FLOAT;
	  else
	    diagptable = tr->partitionData[model].left;


	  if(isTip(pNumber, tr->mxtips) || isTip(qNumber, tr->mxtips))
	    {	        	    
	      if(isTip(qNumber, tr->mxtips))
		{	
		  
		  if(!tr->useFloat)
		    x2_start = tr->partitionData[model].xVector[pNumber - tr->mxtips -1];
		  else
		    x2_start_FLOAT = tr->partitionData[model].xVector_FLOAT[pNumber - tr->mxtips -1];

		  if(!tr->useFastScaling)
		    ex2      = tr->partitionData[model].expVector[pNumber - tr->mxtips - 1];
		  
		  tip = tr->partitionData[model].yVector[qNumber];	 	      
		}           
	      else
		{
		  if(!tr->useFloat)
		    x2_start = tr->partitionData[model].xVector[qNumber - tr->mxtips - 1];
		  else
		    x2_start_FLOAT = tr->partitionData[model].xVector_FLOAT[qNumber - tr->mxtips - 1];

		  if(!tr->useFastScaling)
		    ex2      = tr->partitionData[model].expVector[qNumber - tr->mxtips - 1];	 
		  
		  tip = tr->partitionData[model].yVector[pNumber];
		}
	    }
	  else
	    {  
	      if(!tr->useFloat)               
		{
		  x1_start = tr->partitionData[model].xVector[pNumber - tr->mxtips - 1];
		  x2_start = tr->partitionData[model].xVector[qNumber - tr->mxtips - 1];
		}
	      else
		{
		  x1_start_FLOAT = tr->partitionData[model].xVector_FLOAT[pNumber - tr->mxtips - 1];
		  x2_start_FLOAT = tr->partitionData[model].xVector_FLOAT[qNumber - tr->mxtips - 1];
		}

	      if(!tr->useFastScaling)
		{
		  ex1      = tr->partitionData[model].expVector[pNumber - tr->mxtips - 1];
		  ex2      = tr->partitionData[model].expVector[qNumber - tr->mxtips - 1];     
		}
	    }


	  if(tr->multiBranch)
	    z = pz[model];
	  else
	    z = pz[0];

	  switch(tr->partitionData[model].dataType)
	    { 
	    case BINARY_DATA:
	       switch(tr->rateHetModel)
		{
		case CAT:	    
		  {		   		    
		    calcDiagptable(z, BINARY_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);
		    
		    partitionLikelihood =  evaluateGTRCAT_BINARY(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
								 x1_start, x2_start, tr->partitionData[model].tipVector, 
								 tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;	  	   
		case GAMMA:	   
		  {		    		    
		    calcDiagptable(z, BINARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);		    		    

		    partitionLikelihood = evaluateGTRGAMMA_BINARY(ex1, ex2, tr->partitionData[model].wgt,
								  x1_start, x2_start, tr->partitionData[model].tipVector,
								  tip, width, diagptable, _vector, writeVector, tr->useFastScaling); 		   
		  }
		  break; 
		case GAMMA_I:
		  {		    		    
		    calcDiagptable(z, BINARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRGAMMAINVAR_BINARY(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
								       x1_start, x2_start,
								       tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								       tr->partitionData[model].propInvariant,
								       tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;
		default:
		  assert(0);
		}
	      break;	   
	    case DNA_DATA:
	      switch(tr->rateHetModel)
		{
		case CAT:
		  if(tr->useFloat)
		    {		     		
		      calcDiagptable_FLOAT(z, DNA_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable_FLOAT);
		
		      partitionLikelihood =  evaluateGTRCAT_FLOAT(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
								  x1_start_FLOAT, x2_start_FLOAT, tr->partitionData[model].tipVector_FLOAT, 
								  tip, width, diagptable_FLOAT, _vector_FLOAT, writeVector, tr->useFastScaling);		      
		    }
		  else
		    {
		      calcDiagptable(z, DNA_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);
		      
		      partitionLikelihood =  evaluateGTRCAT(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
							    x1_start, x2_start, tr->partitionData[model].tipVector, 
							    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		      		      
		    }
		  break;	  	   
		case GAMMA:
		  if(tr->useFloat)
		    {
		      calcDiagptable_FLOAT(z, DNA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable_FLOAT);		    		    
		      
		      partitionLikelihood = evaluateGTRGAMMA_FLOAT(ex1, ex2, tr->partitionData[model].wgt,
								   x1_start_FLOAT, x2_start_FLOAT, tr->partitionData[model].tipVector_FLOAT,
								   tip, width, diagptable_FLOAT, _vector_FLOAT, writeVector, tr->useFastScaling); 		      
		    }
		  else
		    {		     		      
		      calcDiagptable(z, DNA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);		    		    
		      
		      partitionLikelihood = evaluateGTRGAMMA(ex1, ex2, tr->partitionData[model].wgt,
							     x1_start, x2_start, tr->partitionData[model].tipVector,
							     tip, width, diagptable, _vector, writeVector, tr->useFastScaling); 		      
		    }
		  break; 
		case GAMMA_I:
		  {
		    calcDiagptable(z, DNA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRGAMMAINVAR(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
								x1_start, x2_start,
								tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								tr->partitionData[model].propInvariant,
								tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;
		default:
		  assert(0);
		}
	      break;
	    case AA_DATA:
	      switch(tr->rateHetModel)
		{
		case CAT:	    
		  {		   
		    calcDiagptable(z, AA_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRCATPROT(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
							     x1_start, x2_start, tr->partitionData[model].tipVector,
							     tip, width, diagptable, _vector, writeVector, tr->useFastScaling);		  
		  }	     	      
		  break;	      
		case GAMMA:
		  if(tr->useFloat)
		    {		      		      
		      calcDiagptable_FLOAT(z, AA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable_FLOAT);
		      
		      partitionLikelihood = evaluateGTRGAMMAPROT_FLOAT(ex1, ex2, tr->partitionData[model].wgt,
								       x1_start_FLOAT, x2_start_FLOAT, tr->partitionData[model].tipVector_FLOAT,
								       tip, width, diagptable_FLOAT, _vector_FLOAT, writeVector, tr->useFastScaling);		      

		    }
		  else
		    {
		      calcDiagptable(z, AA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		      
		      partitionLikelihood = evaluateGTRGAMMAPROT(ex1, ex2, tr->partitionData[model].wgt,
								 x1_start, x2_start, tr->partitionData[model].tipVector,
								 tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		    }
		  break;
		case GAMMA_I:		  	    
		  {
		    calcDiagptable(z, AA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		    
		    partitionLikelihood = evaluateGTRGAMMAPROTINVAR(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
								    x1_start, x2_start, 
								    tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								    tr->partitionData[model].propInvariant, 
								    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	  
		  break;
		default:
		  assert(0);
		}
	      break;
	    case SECONDARY_DATA:
	      switch(tr->rateHetModel)
		{
		case CAT:	    
		  {
		    calcDiagptable(z, SECONDARY_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRCATSECONDARY(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
								  x1_start, x2_start, tr->partitionData[model].tipVector,
								  tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	     	      
		  break;	      
		case GAMMA:
		  {
		    calcDiagptable(z, SECONDARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRGAMMASECONDARY(ex1, ex2, tr->partitionData[model].wgt,
								    x1_start, x2_start, tr->partitionData[model].tipVector,
								    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;
		case GAMMA_I:		  	    
		  {
		    calcDiagptable(z, SECONDARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		    
		    partitionLikelihood = evaluateGTRGAMMASECONDARYINVAR(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
									 x1_start, x2_start, 
									 tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
									 tr->partitionData[model].propInvariant, 
									 tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	  
		  break;
		default:
		  assert(0);
		}
	      break;
	    case SECONDARY_DATA_6:
	      switch(tr->rateHetModel)
		{
		case CAT:	    
		  {
		    calcDiagptable(z, SECONDARY_DATA_6, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRCATSECONDARY_6(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
								    x1_start, x2_start, tr->partitionData[model].tipVector,
								    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	     	      
		  break;	      
		case GAMMA:
		  {
		    calcDiagptable(z, SECONDARY_DATA_6, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRGAMMASECONDARY_6(ex1, ex2, tr->partitionData[model].wgt,
								    x1_start, x2_start, tr->partitionData[model].tipVector,
								    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;
		case GAMMA_I:		  	    
		  {
		    calcDiagptable(z, SECONDARY_DATA_6, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		    
		    partitionLikelihood = evaluateGTRGAMMASECONDARYINVAR_6(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
									   x1_start, x2_start, 
									   tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
									   tr->partitionData[model].propInvariant, 
									   tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	  
		  break;
		default:
		  assert(0);
		}
	      break;
	    case SECONDARY_DATA_7:
	      switch(tr->rateHetModel)
		{
		case CAT:	    
		  {
		    calcDiagptable(z, SECONDARY_DATA_7, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRCATSECONDARY_7(ex1, ex2, tr->partitionData[model].rateCategory, tr->partitionData[model].wgt,
								    x1_start, x2_start, tr->partitionData[model].tipVector,
								    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);		   
		  }	     	      
		  break;	      
		case GAMMA:
		  {
		    calcDiagptable(z, SECONDARY_DATA_7, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);

		    partitionLikelihood = evaluateGTRGAMMASECONDARY_7(ex1, ex2, tr->partitionData[model].wgt,
								      x1_start, x2_start, tr->partitionData[model].tipVector,
								      tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }
		  break;
		case GAMMA_I:		  	    
		  {
		    calcDiagptable(z, SECONDARY_DATA_7, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		    
		    partitionLikelihood = evaluateGTRGAMMASECONDARYINVAR_7(ex1, ex2, tr->partitionData[model].wgt, tr->partitionData[model].invariant,
									   x1_start, x2_start, 
									   tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
									   tr->partitionData[model].propInvariant, 
									   tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		  }	  
		  break;
		default:
		  assert(0);
		}
	      break;
	    default:
	      assert(0);
	    }
	  
	  if(tr->useFastScaling)
	    {
	      if(tr->useFloat)
		partitionLikelihood += (tr->partitionData[model].globalScaler[pNumber] + tr->partitionData[model].globalScaler[qNumber]) * log(minlikelihood_FLOAT);
	      else
		partitionLikelihood += (tr->partitionData[model].globalScaler[pNumber] + tr->partitionData[model].globalScaler[qNumber]) * log(minlikelihood);
	    }
	  
	  result += partitionLikelihood;	  
	  tr->perPartitionLH[model] = partitionLikelihood;
	}            
    }
      
  return result;
}


double evaluateGeneric (tree *tr, nodeptr p)
{
  volatile double result;
  nodeptr q = p->back; 
  int i;
    
  tr->td[0].ti[0].pNumber = p->number;
  tr->td[0].ti[0].qNumber = q->number;          
  
  for(i = 0; i < tr->numBranches; i++)    
    tr->td[0].ti[0].qz[i] =  q->z[i];
  
  tr->td[0].count = 1;
  if(!p->x)
    computeTraversalInfo(p, &(tr->td[0].ti[0]), &(tr->td[0].count), tr->mxtips, tr->numBranches);
  if(!q->x)
    computeTraversalInfo(q, &(tr->td[0].ti[0]), &(tr->td[0].count), tr->mxtips, tr->numBranches);  
  
#ifdef _USE_PTHREADS 
 {
    int j;
    
    masterBarrier(THREAD_EVALUATE, tr); 
    if(tr->NumberOfModels == 1)
      {
	for(i = 0, result = 0.0; i < NumberOfThreads; i++)          
	  result += reductionBuffer[i];  	  	     
      
	tr->perPartitionLH[0] = result;
      }
    else
      {
	volatile double partitionResult;
	
	result = 0.0;

	for(j = 0; j < tr->NumberOfModels; j++)
	  {
	    for(i = 0, partitionResult = 0.0; i < NumberOfThreads; i++)          	      
	      partitionResult += reductionBuffer[i * tr->NumberOfModels + j];
	    result += partitionResult;
	    tr->perPartitionLH[j] = partitionResult;
	  }
      }
  }  
#else
  result = evaluateIterative(tr, FALSE);
#endif   

  tr->likelihood = result;    

  return result;
}

double evaluateGenericInitrav (tree *tr, nodeptr p)
{
  volatile double result;   

  determineFullTraversal(p, tr);
      
#ifdef _USE_PTHREADS 
  {
    int i, j;
    
    masterBarrier(THREAD_EVALUATE, tr);    

    if(tr->NumberOfModels == 1)
      {
	for(i = 0, result = 0.0; i < NumberOfThreads; i++)          
	  result += reductionBuffer[i];  	  	     
      
	tr->perPartitionLH[0] = result;
      }
    else
      {
	volatile double partitionResult;
	
	result = 0.0;

	for(j = 0; j < tr->NumberOfModels; j++)
	  {
	    for(i = 0, partitionResult = 0.0; i < NumberOfThreads; i++)          	      
	      partitionResult += reductionBuffer[i * tr->NumberOfModels + j];
	    result +=  partitionResult;
	    tr->perPartitionLH[j] = partitionResult;
	  }
      }
    
  }
#else
  result = evaluateIterative(tr, FALSE);
#endif


 

  tr->likelihood = result;         

  return result;
}


void onlyInitrav(tree *tr, nodeptr p)
{   
  determineFullTraversal(p, tr);  

#ifdef _USE_PTHREADS  
  masterBarrier(THREAD_NEWVIEW, tr);  	 
#else
  newviewIterative(tr);   
#endif     
}



void evaluateGenericVector (tree *tr, nodeptr p)
{
  nodeptr q = p->back;
  int i;  
    
  assert(isTip(p->number, tr->mxtips)  || isTip(q->number, tr->mxtips));
      
  if(isTip(q->number, tr->mxtips))
    {
      nodeptr tmp = p;
      p = q;
      q = tmp;	      
    }   
  
  tr->td[0].ti[0].pNumber = p->number;
  tr->td[0].ti[0].qNumber = q->number;

  for(i = 0; i < tr->numBranches; i++)    
    tr->td[0].ti[0].qz[i] =  q->z[i];

  tr->td[0].count = 1;
  if(!q->x)
    computeTraversalInfo(q, &(tr->td[0].ti[0]), &(tr->td[0].count), tr->mxtips, tr->numBranches); 
 
#ifdef _USE_PTHREADS
  masterBarrier(THREAD_EVALUATE_VECTOR, tr);
#else
  evaluateIterative(tr, TRUE); 
#endif
}



#ifdef _USE_PTHREADS_MULTIGRAIN

double evalCL(tree *tr, int tipCase, double *x1, double *x2, int *ex1, int *ex2, unsigned char *tip, double pz)
{
  double 
    result = 0.0,
    *diagptable = tr->partitionData[0].left;
  
  int *rateCategory = tr->contiguousRateCategory;
  int *wgt          = tr->contiguousWgt;
  int *invariant    = tr->contiguousInvariant;   
  
  int width = tr->partitionData[0].upper - tr->partitionData[0].lower;

  {        
    calcDiagptable(pz, DNA_DATA, 4, tr->partitionData[0].gammaRates, tr->partitionData[0].EIGN, diagptable);		    		    
    
    result = evaluateGTRGAMMA(ex1, ex2, wgt, x1, x2, tr->partitionData[0].tipVector,
			      tip, width, diagptable, (double*)NULL, FALSE, tr->useFastScaling); 
   
  }

  tr->likelihood = result;

  return result;
}

double evaluateInsert(tree *tr, double *pz, int insertNumber)
{
  double 
    result = 0.0;  
  int model, columnCounter, offsetCounter;
  double *_vector = (double*)NULL;
  boolean writeVector = FALSE;

  for(model = 0, columnCounter = 0, offsetCounter = 0; model < tr->NumberOfModels; model++)
    {                      
      int width = tr->partitionData[model].upper - tr->partitionData[model].lower;
      double z, partitionLikelihood;  
    
      int    
	increment, 
	*ex1 = (int*)NULL, 
	*ex2 = (int*)NULL;

      double 
	*x1_start = (double*)NULL, 
	*x2_start = (double*)NULL,
	*diagptable = tr->partitionData[model].left;

      unsigned char *tip = (unsigned char*)NULL;

      /* get the additional contiguous pointers we need to compute the likelihood */

      int *rateCategory = &tr->contiguousRateCategory[columnCounter];
      int *wgt          = &tr->contiguousWgt[columnCounter];
      int *invariant    = &tr->contiguousInvariant[columnCounter];           

      ex2      = &tr->temporaryScaling[columnCounter];
      x2_start = &tr->temporaryVector[offsetCounter];
      tip      = &tr->contiguousTips[insertNumber][columnCounter];
	
      /* no need to worry about this now */

      if(tr->multiBranch)
	z = pz[model];
      else
	z = pz[0];
	
      switch(tr->partitionData[model].dataType)
	{ 	 
	case BINARY_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:	    
	      {					
		calcDiagptable(z, BINARY_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);				
		
		partitionLikelihood =  evaluateGTRCAT_BINARY(ex1, ex2, rateCategory, wgt,
							     x1_start, x2_start, tr->partitionData[model].tipVector, 
							     tip, width, diagptable, _vector, writeVector, tr->useFastScaling);	
		increment = 2;	       
	      }
	      break;	  	   
	    case GAMMA:	   
	      {			
		calcDiagptable(z, BINARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);		    		    
		
		partitionLikelihood = evaluateGTRGAMMA_BINARY(ex1, ex2, wgt,
							      x1_start, x2_start, tr->partitionData[model].tipVector,
							      tip, width, diagptable, _vector, writeVector, tr->useFastScaling); 
		increment = 8;	
	      }
	      break; 
	    case GAMMA_I:
	      {			
		calcDiagptable(z, BINARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMAINVAR_BINARY(ex1, ex2, wgt, invariant,
								   x1_start, x2_start,
								   tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								   tr->partitionData[model].propInvariant,
								   tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 8;	
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case DNA_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:	    
	      {
		calcDiagptable(z, DNA_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood =  evaluateGTRCAT(ex1, ex2, rateCategory, wgt,
						      x1_start, x2_start, tr->partitionData[model].tipVector, 
						      tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		
		increment = 4;	       
	      }
	      break;	  	   
	    case GAMMA:	   
	      {			
		calcDiagptable(z, DNA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);		    		    
		
		partitionLikelihood = evaluateGTRGAMMA(ex1, ex2, wgt,
						       x1_start, x2_start, tr->partitionData[model].tipVector,
						       tip, width, diagptable, _vector, writeVector, tr->useFastScaling); 
		increment = 16;		
	      }
	      break; 
	    case GAMMA_I:
	      {
		calcDiagptable(z, DNA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMAINVAR(ex1, ex2, wgt, invariant,
							    x1_start, x2_start,
							    tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
							    tr->partitionData[model].propInvariant,
							    tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 16;		
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case AA_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:	    
	      {				
		calcDiagptable(z, AA_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRCATPROT(ex1, ex2, rateCategory, wgt,
							 x1_start, x2_start, tr->partitionData[model].tipVector,
							 tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 20;	      
	      }	     	      
	      break;	      
	    case GAMMA:
	      {
		calcDiagptable(z, AA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMAPROT(ex1, ex2, wgt,
							   x1_start, x2_start, tr->partitionData[model].tipVector,
							   tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 80;				    
	      }
	      break;
	    case GAMMA_I:		  	    
	      {
		calcDiagptable(z, AA_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMAPROTINVAR(ex1, ex2, wgt, invariant,
								x1_start, x2_start, 
								tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								tr->partitionData[model].propInvariant, 
								tip, width, diagptable, _vector, writeVector, tr->useFastScaling); 
		increment = 80;	       
	      }	  
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case SECONDARY_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:	    
	      {
		calcDiagptable(z, SECONDARY_DATA, tr->NumberOfCategories, tr->cdta->patrat, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRCATSECONDARY(ex1, ex2, rateCategory, wgt,
							      x1_start, x2_start, tr->partitionData[model].tipVector,
							      tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 16;		
	      }	     	      
	      break;	      
	    case GAMMA:
	      {			
		calcDiagptable(z, SECONDARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMASECONDARY(ex1, ex2, wgt,
								x1_start, x2_start, tr->partitionData[model].tipVector,
								tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 64;				    
	      }
	      break;
	    case GAMMA_I:		  	    
	      {
		calcDiagptable(z, SECONDARY_DATA, 4, tr->partitionData[model].gammaRates, tr->partitionData[model].EIGN, diagptable);
		
		partitionLikelihood = evaluateGTRGAMMASECONDARYINVAR(ex1, ex2, wgt, invariant,
								     x1_start, x2_start, 
								     tr->partitionData[model].tipVector, tr->partitionData[model].frequencies, 
								     tr->partitionData[model].propInvariant, 
								     tip, width, diagptable, _vector, writeVector, tr->useFastScaling);
		increment = 64;	      
	      }	  
	      break;
	    default:
	      assert(0);
	    }
	  break;
	default:
	  assert(0);
	}
     
      result += partitionLikelihood;	  
      tr->perPartitionLH[model] = partitionLikelihood;      

      columnCounter += width;
      offsetCounter += width * increment;
    }  

  return result;
}



#endif




